using System.Console;
using Oyster.Units;
using Oyster.Units.Macros;

def m1 = Cgs.Mass(1); // 1 gram
def m2 = Si.Mass(m1); // Automatically converted into kilograms. Conversion result is 0.001 kg

WriteLine($"Mass in SI: $m2, in CGS: $m1");

def l1 = Cgs.Length(100); // 100 cm
def l2 = Si.Length(l1);   // Converted automatically. Result is 1 m

WriteLine($"Length in SI: $l2, in CGS: $l1");

def t1 = Cgs.Time(1); // 1 sec
def t2 = Si.Time(t1); // 1 sec as well because time measurement units are the same in SI and CGS

WriteLine($"Time in SI: $t2, in CGS: $t1");

// Here is the more complex example. Previous units (mass, length, time) are base units
// in both systems of units. But here we are working with derived units -
// velocity, energy and power
def v1 = Cgs.Velocity(10); // 10 cm/sec
def v2 = Si.Velocity(v1);  // Conversion result - 0.1 m/sec

WriteLine($"Velocity in SI: $v2, in CGS: $v1");

def e1 = Si.Energy(0.0001); // 10e-4 J
def e2 = Cgs.Energy(e1);    // e2 = 1000 erg

WriteLine($"Energy in SI: $e1, in CGS: $e2");

// This is example of working with non-system units. Here power is converted from watts
// into horse powers
def p1 = Si.Power(1000); // 1000 W
def p2 = HorsePower(p1);  // 1.35962 hp

WriteLine($"Power in SI: $p1, in Ext system: $p2");

// Example of simple ariphmetic operations
def a1 = Si.Acceleration(1); // 1 m/sec^2
def a2 = Cgs.Acceleration(a1 + Cgs.Acceleration(200)); // 1 m/sec^2 + 200 cm/sec^2 = 300 cm/sec^2 - everything's correct

WriteLine($"Acceleration: a1 = $a1, a2 = $a2");

// Force
def f1 = Si.Force(10);      // 10 N
def f2 = Cgs.Force(f1*2);  // 10 N * 2 = 2000000 dyn

def f3 : Si.Force = m1*l2/(t1*t1); // 1 g * 1 m / (1 sec * 1 sec) = 0.001 N
// This won't compile - wrong measure
//def f4 = Si.Force(m1*l2/t1);
// That's an error message from ncc:
//  Main.n(52,10,52,18): error : each overload has an error during call:
//  Main.n(52,10,52,18): error : overload #1, method Oyster.Units.Si.Force..ctor(value : System.Double) : void
//  Main.n(52,10,52,18): error : in argument #1 (value) of Oyster.Units.Si.Force[Oyster.Units.IUnit_1_1_-2_0_0_0_0+], needed a System.Double, got _N_Unit3121[Oyster.Units.IUnit_1_1_-1_0_0_0_0+]: _N_Unit3121 is not a subtype of System.Double [simple require]
//  Main.n(52,10,52,18): error : overload #2, method Oyster.Units.Si.Force..ctor(unit : 'a.235) : void
//  Main.n(52,10,52,18): error : in argument #1 (unit) of Oyster.Units.Si.Force[(Oyster.Units.IUnit_1_1_-2_0_0_0_0 TILL _N_Unit3121[Oyster.Units.IUnit_1_1_-1_0_0_0_0+])], needed a Oyster.Units.IUnit_1_1_-2_0_0_0_0+, got _N_Unit3121[Oyster.Units.IUnit_1_1_-1_0_0_0_0+]: _N_Unit3121 is not a subtype of Oyster.Units.IUnit_1_1_-2_0_0_0_0 [simple require]
//  Main.n(52,10,52,18): error : overload #3, Oyster.Units.Si.Force
//  Main.n(52,10,52,18): error : wrong number of parameters in call, needed 0, got 1

WriteLine($"Force: f1 = $f1, f2 = $f2, f3 = $f3");

// Compile-time error here - trying to add energy, mass and velocity
//WriteLine($"ERROR = $(e1 + e2 + v1 + v2 + m1 + m2)");
// This is an error from ncc (from my macro, to be precise):
//  Main.n(65,52,65,64): error : Expression { e1 + e2 [+] v1 } can't be compiled because units are having different dimensions.
//  Main.n(65,52,65,64): error : Expression { e1 + e2 [+] v1 } can't be compiled because units are having different dimensions.
//  Main.n(65,52,65,74): error : Expression { e1 + e2 + v1 + v2 [+] m1 } can't be compiled because units are having different dimensions.
//  Main.n(65,52,65,64): error : Expression { e1 + e2 [+] v1 } can't be compiled because units are having different dimensions.
//  Main.n(65,52,65,64): error : Expression { e1 + e2 [+] v1 } can't be compiled because units are having different dimensions.
//  Main.n(65,52,65,74): error : Expression { e1 + e2 + v1 + v2 [+] m1 } can't be compiled because units are having different dimensions.

// Here conversion is performed "on the fly". Result is 0.0002 J - it has the same type as left operand
WriteLine($"Result = $(e1 + e2)");

// Here 1 is written because e1 and e2 both contain the same value but in different systems of units
WriteLine($"Result = $(e1/e2)");

// Physical literals usage
def m3 = 1 g;
def m4 = Si.Mass(m1);

WriteLine($"Mass in SI: $m4, in CGS: $m3");

def x1 = Si.Area(1 cm * 10 m);

WriteLine($"Area of 1 cm * 10 m = $x1 m")
